var EXPORTED_SYMBOLS = ["prefManager"];

prefManager = {
  srv: Components.classes["@mozilla.org/preferences;1"].getService(Components.interfaces.nsIPrefBranch2),
  setBranch: function(name) { this.srv = this.prefService.getBranch(name); },

  getBool: function(name) {return this.srv.getBoolPref(name);},
  setBool: function(name, value) {this.srv.setBoolPref(name, value);},

  getString: function(name) {return this.srv.getCharPref(name);},
  setString: function(name, value) {this.srv.setCharPref(name, value);},

  getInt: function(name) {return this.srv.getIntPref(name);},
  setInt: function(name, value) {this.srv.setIntPref(name, value);},
  deleteBranch: function(name) {this.srv.deleteBranch(name);},
  hasUserValue: function(name) {this.srv.prefHasUserValue(name);},

  registerChangeCallback: function(callback, branch, notifyNow){
    if (!this.callbacks[branch]) {
      this.callbacks[branch] = [callback];
      this.srv.addObserver(branch, this, false);
    } else if (this.callbacks[branch].indexOf(callback) < 0)
      this.callbacks[branch].push(callback);

    if (!notifyNow)
      return;

    var list = this.srv.getChildList(branch, {});

    for (var i = 0; i < list.length; i++)
      callback(list[i], this.getPref(list[i]));
  },

  unregisterChangeCallback: function(callback, branch){
    if (branch != null) {
      var r = {};
      r[branch] = 1;
      branch = r;
    } else
      branch = Iterator(this.callbacks, true);

    for (i in branch) {
      var idx = this.callbacks[i] && this.callbacks[i].indexOf(callback);
      if (idx != null && idx >= 0) {
        this.callbacks[i].splice(idx, 1);
        if (this.callbacks[i].length == 0) {
            this.srv.removeObserver(i, this)
            delete this.callbacks[i];
        }
      }
    }
  },

  getPref: function(name) {
    try {
      var type = this.srv.getPrefType(name);
      var value = 
        type == this.srv.PREF_BOOL ? this.srv.getBoolPref(name) :
        type == this.srv.PREF_INT ? this.srv.getIntPref(name) :
        type == this.srv.PREF_STRING ? this.srv.getComplexValue(name, Components.interfaces.nsISupportsString).data :
        null;
      return value;
    } catch(ex) {
        return null;
    }
  },
  setPref: function(name, value) {
    var valueType = typeof(value);
    if(valueType == "boolean")
      this.srv.setBoolPref(name, value);
    if(valueType == "number")
      this.srv.setIntPref(name, value);
    if(valueType == "string"){
      var str = Components.classes["@mozilla.org/supports-string;1"].createInstance(Components.interfaces.nsISupportsString);
      str.data = value;
      this.srv.setComplexValue(name, Components.interfaces.nsISupportsString, str);
    }
  },
  resetPref: function(name) {
    try {
      this.srv.clearUserPref(name);
    } catch(ex) { }
  },
}
